import abc

import typing

from hpotk.model import TermId

NODE = typing.TypeVar('NODE', bound=TermId)
# Term ID that is added as an artificial root if >1 root candidates are found in the ontology graph.
OWL_THING = TermId.from_curie("owl:Thing")


class OntologyGraph(typing.Generic[NODE], metaclass=abc.ABCMeta):
    """
    A simple graph with one node type and one edge type.

    The graph is generic over a node type which must extend `hpotk.base.model.TermId`. The graph must not be empty
    """

    @property
    @abc.abstractmethod
    def root(self) -> NODE:
        pass

    @abc.abstractmethod
    def get_children(self, source: NODE) -> typing.Iterator[NODE]:
        pass

    @abc.abstractmethod
    def get_parents(self, source: NODE) -> typing.Iterator[NODE]:
        pass

    @abc.abstractmethod
    def __contains__(self, item: NODE) -> bool:
        pass

    @abc.abstractmethod
    def __iter__(self) -> typing.Iterator[NODE]:
        pass


class GraphAware(typing.Generic[NODE], metaclass=abc.ABCMeta):
    """
    Base class for entities that have an `OntologyGraph`.
    """

    @property
    @abc.abstractmethod
    def graph(self) -> OntologyGraph[NODE]:
        """
        :return: the ontology graph with nodes of a given type.
        """
        pass



import os
import unittest

import ddt
from pkg_resources import resource_filename

from hpotk.model import MinimalTerm, TermId
from hpotk.ontology.load.obographs import load_minimal_ontology
from hpotk.validate import ValidationResult, ValidationLevel
from hpotk.validate import AnnotationPropagationValidator, PhenotypicAbnormalityValidator, ObsoleteTermIdsValidator


TOY_HPO = resource_filename(__name__, os.path.join('data', 'hp.toy.json'))


class TestBaseRuleValidator(unittest.TestCase):

    o = None
    example_terms = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.o = load_minimal_ontology(TOY_HPO)
        cls.example_terms = [
            MinimalTerm.create_minimal_term(
                TermId.from_curie('HP:0001166'),
                name='Arachnodactyly', alt_term_ids=[], is_obsolete=False),
            MinimalTerm.create_minimal_term(
                TermId.from_curie('HP:0002266'),
                name='Focal clonic seizure', alt_term_ids=[], is_obsolete=False),
            MinimalTerm.create_minimal_term(
                TermId.from_curie('HP:0032648'),
                name='Tubularization of Bowman capsule', alt_term_ids=[], is_obsolete=False)
        ]


@ddt.ddt
class TestAnnotationPropagationValidator(TestBaseRuleValidator):

    def setUp(self) -> None:
        self.validator = AnnotationPropagationValidator(self.o)

    def test_empty_input_is_allowed(self):
        results = self.validator.validate([])
        self.assertTrue(results.is_ok())

    def test_ok_input_produces_no_errors(self):
        results = self.validator.validate(self.example_terms)
        self.assertTrue(results.is_ok())

    @ddt.data(
        ("HP:0100807", "HP:0100807", "Long fingers"),
        ("HP:0006010", 'HP:0100807', "Long fingers"),
    )
    @ddt.unpack
    def test_ancestor_presence_produces_error(self, query_id, current_id, name):
        example_terms = list(self.example_terms)
        example_terms.append(
            MinimalTerm.create_minimal_term(
                TermId.from_curie(query_id), name=name, alt_term_ids=[], is_obsolete=False)
        )
        results = self.validator.validate(example_terms)
        self.assertFalse(results.is_ok())
        self.assertEqual(results.results[0],
                         ValidationResult(level=ValidationLevel.ERROR, category='annotation_propagation',
                                          message='Terms should not contain both Arachnodactyly [HP:0001166] '
                                                  f'and its ancestor {name} [{current_id}]'))


@ddt.ddt
class TestPhenotypicAbnormalityValidator(TestBaseRuleValidator):

    def setUp(self) -> None:
        self.validator = PhenotypicAbnormalityValidator(self.o)

    def test_ok_input_produces_no_errors(self):
        results = self.validator.validate(self.example_terms)
        self.assertTrue(results.is_ok())

    @ddt.data(
        ("HP:0012823", "HP:0012823", "Clinical modifier"),
        ("HP:0003825", 'HP:0003828', "Variable expressivity"),
        ("HP:0003621", 'HP:0003621', "Juvenile onset"),
    )
    @ddt.unpack
    def test_clinical_modifier_presence_produces_error(self, query_id, current_id, name):
        example_terms = list(self.example_terms)
        example_terms.append(
            MinimalTerm.create_minimal_term(
                TermId.from_curie(query_id), name=name, alt_term_ids=[], is_obsolete=False)
        )
        results = self.validator.validate(example_terms)
        self.assertFalse(results.is_ok())
        self.assertEqual(results.results[0],
                         ValidationResult(level=ValidationLevel.ERROR, category='phenotypic_abnormality_descendant',
                                          message=f'{name} [{current_id}] is not a descendant of '
                                                  'Phenotypic abnormality [HP:0000118]'))


@ddt.ddt
class TestObsoleteTermIdsValidator(TestBaseRuleValidator):

    def setUp(self) -> None:
        self.validator = ObsoleteTermIdsValidator(self.o)

    def test_ok_input_produces_no_errors(self):
        results = self.validator.validate(self.example_terms)
        self.assertTrue(results.is_ok())

    @ddt.data(
        ("HP:0006010", 'HP:0100807', "Long fingers"),
    )
    @ddt.unpack
    def test_clinical_modifier_presence_produces_error(self, query_id, current_id, name):
        example_terms = list(self.example_terms)
        example_terms.append(
            MinimalTerm.create_minimal_term(
                TermId.from_curie(query_id), name=name, alt_term_ids=[], is_obsolete=False)
        )
        results = self.validator.validate(example_terms)
        self.assertFalse(results.is_ok())
        self.assertEqual(results.results[0],
                         ValidationResult(level=ValidationLevel.WARNING, category='obsolete_term_id_is_used',
                                          message=f'Using the obsolete {query_id} instead of {current_id} for {name}'))



import unittest

import ddt

from hpotk.model import *


@ddt.ddt
class TestTermId(unittest.TestCase):

    @ddt.data("HP:1234567", "HP_1234567")
    def test_from_curie(self, curie):
        term_id = TermId.from_curie(curie)
        assert term_id.value == "HP:1234567"

    @ddt.data(
        ("HP1234567", "The CURIE HP1234567 has no colon `:` or underscore `_`"),
        (None, "Curie must not be None")
    )
    @ddt.unpack
    def test_from_curie__failures(self, curie, message):
        with self.assertRaises(ValueError) as cm:
            TermId.from_curie(curie)
        self.assertEqual(message, cm.exception.args[0])

    @ddt.unpack
    @ddt.data(
        ["HP:1234567", "HP", "1234567"],
        ["HP_1234567", "HP", "1234567"]
    )
    def test_properties(self, curie, prefix, id):
        term_id = TermId.from_curie(curie)
        self.assertEqual(prefix, term_id.prefix)
        self.assertEqual(id, term_id.id)
        self.assertEqual(f'{prefix}:{id}', term_id.value)

    @ddt.unpack
    @ddt.data(
        ["HP:1234567", "HP:1234567"],
        ["HP:1234567", "HP_1234567"],
        ["HP_1234567", "HP:1234567"],
        ["HP_1234567", "HP_1234567"],
    )
    def test_equality(self, left, right):
        left = TermId.from_curie(left)
        right = TermId.from_curie(right)
        self.assertEqual(left, right)

    @ddt.unpack
    @ddt.data(
        # First compare by prefix
        ["AA:0000000", "BB:0000000", False, False, True],
        ["BB:0000000", "BB:0000000", False, True, False],
        ["CC:0000000", "BB:0000000", True, False, False],
        # Then by value
        ["HP:0000001", "HP:0000002", False, False, True],
        ["HP:0000002", "HP:0000002", False, True, False],
        ["HP:0000003", "HP:0000002", True, False, False],
    )
    def test_comparison(self, left, right, is_gt, is_eq, is_lt):
        left = TermId.from_curie(left)
        right = TermId.from_curie(right)
        self.assertEqual(left > right, is_gt)
        self.assertEqual(left == right, is_eq)
        self.assertEqual(left < right, is_lt)



import unittest

from hpotk.model import *


class TestMinimalTerm(unittest.TestCase):

    def test_equal_terms_are_equal(self):
        one = MinimalTerm.create_minimal_term(TermId.from_curie("HP:1234567"),
                                              "First",
                                              [TermId.from_curie("HP:1111111")],
                                              False)
        other = MinimalTerm.create_minimal_term(TermId.from_curie("HP:1234567"),
                                                "First",
                                                [TermId.from_curie("HP:1111111")],
                                                False)
        self.assertEqual(one, other)

    def test_unequal_terms_are_not_equal(self):
        one = MinimalTerm.create_minimal_term(TermId.from_curie("HP:1234567"),
                                              "First",
                                              [TermId.from_curie("HP:1111111")],
                                              False)
        other = MinimalTerm.create_minimal_term(TermId.from_curie("HP:1234567"),
                                                "Second",
                                                [TermId.from_curie("HP:1111111")],
                                                False)
        self.assertNotEqual(one, other)


class TestTerm(unittest.TestCase):

    def test_equal_terms_are_equal(self):
        one_id = TermId.from_curie("HP:1234567")
        one = Term.create_term(identifier=one_id,
                               name="First",
                               alt_term_ids=[TermId.from_curie("HP:1111111")],
                               is_obsolete=False,
                               definition="First term definition",
                               comment="First comment")
        two_id = TermId.from_curie("HP:1234567")
        two = Term.create_term(identifier=two_id,
                               name="First",
                               alt_term_ids=[TermId.from_curie("HP:1111111")],
                               is_obsolete=False,
                               definition="First term definition",
                               comment="First comment")
        self.assertEqual(one, two)

    def test_not_equal_terms_are_not_equal(self):
        one_id = TermId.from_curie("HP:1111111")
        one = Term.create_term(identifier=one_id,
                               name="First",
                               alt_term_ids=[TermId.from_curie("HP:2222222")],
                               is_obsolete=False,
                               definition="First term definition",
                               comment="Other comment")
        two_id = TermId.from_curie("HP:1111111")
        two = Term.create_term(identifier=two_id,
                               name="First",
                               alt_term_ids=[TermId.from_curie("HP:2222222")],
                               is_obsolete=False,
                               definition="First term definition",
                               comment="First comment")
        self.assertNotEqual(one, two)



import abc
import typing

from hpotk.model import TermId, Term, MinimalTerm

from ._model import Node

MINIMAL_TERM = typing.TypeVar('MINIMAL_TERM', bound=MinimalTerm)


def create_alt_term_ids(node: Node) -> typing.List[TermId]:
    alt_term_ids = []
    if node.meta:
        for bpv in node.meta.basic_property_values:
            if bpv.pred is not None \
                    and bpv.val is not None \
                    and bpv.pred.endswith('#hasAlternativeId'):
                alt_term_ids.append(TermId.from_curie(bpv.val))
    return alt_term_ids


class ObographsTermFactory(typing.Generic[MINIMAL_TERM], metaclass=abc.ABCMeta):
    """
    Term factory turns `TermId` and obographs `Node` into an ontology term.
    """

    @abc.abstractmethod
    def create_term(self, term_id: TermId, node: Node) -> typing.Optional[MINIMAL_TERM]:
        """
        Create `MinimalTerm` or a more specific instance for `TermId` and `Node`

        The term may not be created at the discretion of the factory, in which case `None` is returned.
        """
        pass


class MinimalTermFactory(ObographsTermFactory[MinimalTerm]):

    def create_term(self, term_id: TermId, node: Node) -> typing.Optional[MinimalTerm]:
        is_obsolete = node.meta is not None and node.meta.is_deprecated
        alt_term_ids = create_alt_term_ids(node)
        return MinimalTerm.create_minimal_term(term_id, node.lbl, alt_term_ids, is_obsolete)


class TermFactory(ObographsTermFactory[Term]):

    def create_term(self, term_id: TermId, node: Node) -> typing.Optional[Term]:
        if node.meta:
            definition = node.meta.definition.val if node.meta.definition else None
            comment = ', '.join(node.meta.comments) if len(node.meta.comments) > 0 else None
            alt_term_ids = create_alt_term_ids(node)

            return Term.create_term(term_id, name=node.lbl, alt_term_ids=alt_term_ids,
                                    is_obsolete=node.meta.is_deprecated, definition=definition, comment=comment)
        else:
            return Term.create_term(term_id, name=node.lbl, alt_term_ids=[], is_obsolete=False, definition=None,
                                    comment=None)



import typing
import abc
import enum

from collections import namedtuple

from hpotk.model import Identified


class ValidationLevel(enum.Enum):
    WARNING = 1
    ERROR = 2


ValidationResult = namedtuple('ValidationResult', field_names=['level', 'category', 'message'])


class ValidationResults:
    """
    Results of a single validation.
    """

    def __init__(self, results: typing.Sequence[ValidationResult]):
        self._results = results
        
    @property
    def results(self) -> typing.Sequence[ValidationResult]:
        return self._results

    def is_ok(self) -> bool:
        return len(self._results) == 0

    def __str__(self):
        return f"ValidationResults(is_ok={self.is_ok()}, n_results={len(self._results)})"

    def __repr__(self) -> str:
        return f"ValidationResults(results={[self._results]})"


class RuleValidator(metaclass=abc.ABCMeta):
    """
    `RuleValidator` checks if a sequence of `Identified` items meet the validation requirements.
    The issues are returned as `ValidationResults`.
    """

    @abc.abstractmethod
    def validate(self, items: typing.Sequence[Identified]) -> ValidationResults:
        pass


class ValidationRunner:
    """
    The runner applies a sequence of `RuleValidator`s on items and returns the results as `ValidationResults`.
    """

    def __init__(self, validators: typing.Sequence[RuleValidator]):
        self._validators = validators

    def validate_all(self, items: typing.Sequence[Identified]) -> ValidationResults:
        overall = []
        for validator in self._validators:
            results = validator.validate(items)
            for result in results.results:
                overall.append(result)

        return ValidationResults(overall)



from ._base import Ratio, EvidenceCode, Sex, AnnotationReference
from ._base import HpoDiseaseAnnotation, HpoDisease, HpoDiseases
from ._base import SimpleRatio

from ._simple import SimpleHpoDiseaseAnnotation, SimpleHpoDisease, SimpleHpoDiseases

from . import load



import os
import unittest

from pkg_resources import resource_filename

import hpotk as hp
from hpotk.model import TermId
from hpotk.ontology.load.obographs import *

TOY_HPO = resource_filename(__name__, os.path.join('data', 'hp.toy.json'))

hp.util.setup_logging()


class TestObographs(unittest.TestCase):

    def test_load_minimal_ontology(self):
        o: hp.ontology.MinimalOntology = load_minimal_ontology(TOY_HPO)
        self.assertIsNotNone(o, "Ontology must not be None")
        self.assertIsInstance(o, hp.ontology.MinimalOntology)
        self.assertEqual(o.version, '2022-10-05')
        self.assertEqual(393, len(o), "There must be 393 terms in the ontology")
        self.assertEqual(557, len(list(o.term_ids)), "There must be 557 term IDs in the ontology")
        self.assertEqual(557, len(set(o.term_ids)), "There must be 557 unique term IDs in the ontology")
        self.assertTrue(all([term_id in o for term_id in o.term_ids]), "The ontology must contain all term IDs")
        self.assertTrue(all([o.get_term(k) is not None for k in o.term_ids]),
                        "The `get_term` must get primary term for any term ID from ontology")
        self.assertTrue(all([o.get_term(k.value) is not None for k in o.term_ids]),
                        "The `get_term` must get primary term for any term ID value from ontology")
        self.assertTrue(all([o.get_term(k).identifier == k or k in o.get_term(k).alt_term_ids for k in o.term_ids]),
                        "Each term ID must be either primary or alternative ID")

    def test_load_ontology(self):
        o: hp.ontology.Ontology = load_ontology(TOY_HPO)
        self.assertIsNotNone(o, "Ontology must not be None")
        self.assertIsInstance(o, hp.ontology.Ontology)
        self.assertEqual(o.version, '2022-10-05')
        self.assertEqual(393, len(o), "There must be 393 terms in the ontology")
        self.assertEqual(557, len(list(o.term_ids)), "There must be 557 term IDs in the ontology")
        self.assertEqual(557, len(set(o.term_ids)), "There must be 557 unique term IDs in the ontology")
        self.assertTrue(all([term_id in o for term_id in o.term_ids]), "The ontology must contain all term IDs")
        self.assertTrue(all([o.get_term(k) is not None for k in o.term_ids]),
                        "The `get_term` must get primary term for any term ID from ontology")
        self.assertTrue(all([o.get_term(k).identifier == k or k in o.get_term(k).alt_term_ids for k in o.term_ids]),
                        "Each term ID must be either primary or alternative ID")

    def test_load_minimal_ontology_backed_by_csr(self):
        term_factory = hp.ontology.load.obographs.MinimalTermFactory()
        graph_factory = hp.graph.CsrGraphFactory()
        o: hp.ontology.MinimalOntology = load_minimal_ontology(TOY_HPO, term_factory=term_factory, graph_factory=graph_factory)
        self.assertIsNotNone(o, "Ontology must not be None")

        arachnodactyly = TermId.from_curie("HP:0001166")
        assert all([val.value in {"HP:0001238", "HP:0100807"} for val in (o.graph.get_parents(arachnodactyly))])
        assert len(list(o.graph.get_children(arachnodactyly))) == 0

    @unittest.skip
    def test_print_stats(self):
        import json
        with open(TOY_HPO) as fh:
            graphs = json.load(fh)

        graph = graphs['graphs'][0]
        all_nodes = graph['nodes']
        all_edges = graph['edges']
        print(f'All nodes: {len(all_nodes)}, all edges: {len(all_edges)}')
        current_nodes = self._get_current_nodes(all_nodes)
        print(f'Current nodes: {len(current_nodes)}')
        # Getting the number of all term IDs is too complicated to be implemented here at this moment.
        # The functionality should be implemented later if necessary.

    @staticmethod
    def _get_current_nodes(nodes):
        result = []
        for node in nodes:
            if 'meta' in node:
                meta = node['meta']
                if 'deprecated' in meta:
                    deprecated = meta['deprecated']
                    if not deprecated:
                        result.append(node)
                else:
                    result.append(node)
        return result



__version__ = "0.1.3post0"

from . import model
from . import constants
from . import graph
from . import ontology
from . import algorithm
from . import annotations
from . import validate
from . import util



import typing

from ._term_id import TermId
from ._base import Identified, Named, Versioned
from ._term import MinimalTerm, Term

# Types
ID = typing.TypeVar('ID', bound=TermId)
CURIE_OR_TERM_ID = typing.Union[str, ID]
MINIMAL_TERM = typing.TypeVar('MINIMAL_TERM', bound=MinimalTerm)
TERM = typing.TypeVar('TERM', bound=Term)



import typing

from ._api import OntologyGraph, NODE
from .csr import ImmutableCsrMatrix


class SimpleCsrOntologyGraph(OntologyGraph):
    """
    Simple CSR ontology graph uses a dict to map a NODE to connectivity matrix index.
    """

    # TODO - try replacing `self._node_to_idx` with binary search.
    CHILD_RELATIONSHIP_CODE = 1
    PARENT_RELATIONSHIP_CODE = -1

    def __init__(self, root: NODE,
                 nodes: typing.Sequence[NODE],
                 connectivity_matrix: ImmutableCsrMatrix):
        self._root = root
        self._nodes = nodes
        self._node_to_idx = {node: idx for idx, node in enumerate(nodes)}
        self._connectivity_matrix = connectivity_matrix

    @property
    def root(self) -> NODE:
        return self._root

    def get_children(self, source: NODE) -> typing.Iterator[NODE]:
        return self._get_nodes_with_relationship(source,
                                                 SimpleCsrOntologyGraph.CHILD_RELATIONSHIP_CODE)

    def get_parents(self, source: NODE) -> typing.Iterator[NODE]:
        return self._get_nodes_with_relationship(source,
                                                 SimpleCsrOntologyGraph.PARENT_RELATIONSHIP_CODE)

    def _get_nodes_with_relationship(self, source, relationship):
        node_idx = self._node_to_idx[source]
        child_idxs = self._connectivity_matrix.col_indices_of_val(node_idx, relationship)
        return self._nodes[child_idxs]

    def __contains__(self, item: NODE) -> bool:
        return item in self._node_to_idx

    def __iter__(self) -> typing.Iterator[NODE]:
        return iter(self._nodes)



import abc
import enum
import typing

from hpotk.model import Identified, Named, Versioned, TermId
from hpotk.model import CURIE_OR_TERM_ID


class Ratio(metaclass=abc.ABCMeta):

    @staticmethod
    def create(numerator: int, denominator: int):
        return SimpleRatio(numerator, denominator)

    @property
    @abc.abstractmethod
    def numerator(self) -> int:
        pass

    @property
    @abc.abstractmethod
    def denominator(self) -> int:
        pass

    @property
    def frequency(self) -> float:
        return self.numerator / self.denominator

    def is_positive(self) -> bool:
        return self.numerator > 0

    def is_zero(self) -> bool:
        return self.numerator == 0

    @staticmethod
    def fold(left, right):
        """
        Fold two :class:`Ratio`s together into a new :class:`Ratio` that represents `n` over `m` of both inputs.

        Note that this is *NOT* the addition of two :class:`Ratio`s!

        For instance, if :math:`n_1` of :math:`m_1` and :math:`n_2` of :math:`m_2` population members like lasagna,
        then :math:`n_1 + n_2` of :math:`m_1 + m_2` people like lasagna in total.

        :param left: left :class:`Ratio`.
        :param right: right :class:`Ratio`.
        :return: the result as a :class:`SimpleRatio`.
        """
        if isinstance(left, Ratio) and isinstance(right, Ratio):
            return SimpleRatio(left.numerator + right.numerator, left.denominator + right.denominator)
        else:
            msg = 'left arg must be an instance of `Ratio`' \
                if isinstance(right, Ratio) \
                else 'right arg must be an instance of `Ratio`'
            raise ValueError(msg)

    def __eq__(self, other):
        return isinstance(other, Ratio) \
            and self.numerator * other.denominator == other.numerator * self.denominator

    def __str__(self):
        return f"{self.numerator}/{self.denominator}"


class SimpleRatio(Ratio):

    def __init__(self, numerator: int, denominator: int):
        if not isinstance(numerator, int) or numerator < 0:
            raise ValueError(f'Numerator {numerator} must be a non-negative `int`')
        if not isinstance(denominator, int) or denominator <= 0:
            raise ValueError(f'Denominator {denominator} must be a positive `int`')
        self._numerator = numerator
        self._denominator = denominator

    @property
    def numerator(self) -> int:
        return self._numerator

    @property
    def denominator(self) -> int:
        return self._denominator

    def __eq__(self, other):
        if isinstance(other, SimpleRatio):
            return self._numerator * other._denominator == self._denominator * other._numerator
        else:
            return super().__eq__(other)

    def __repr__(self):
        return f"SimpleRatio(" \
               f"numerator={self._numerator}, " \
               f"denominator={self._denominator})"


class EvidenceCode(enum.Enum):
    """Inferred from electronic evidence."""
    IEA = 0
    """Traceable author statement."""
    TAS = 1
    """Published clinical study."""
    PCS = 2

    @staticmethod
    def parse(value: str):
        """
        Parse evidence code from `str` value.

        :param value: a str with the evidence code.
        :return: the parsed enum member or `None` if `value` is not valid `EvidenceCode` value.
        """
        value = value.upper()
        if value == 'IEA':
            return EvidenceCode.IEA
        elif value == 'TAS':
            return EvidenceCode.TAS
        elif value == 'PCS':
            return EvidenceCode.PCS
        else:
            return None


class Sex(enum.Enum):
    UNKNOWN = 0
    MALE = 1
    FEMALE = 2

    @staticmethod
    def parse(value: str):
        """
        Parse :class:`Sex` from `str` value.

        :param value: a `str` with the sex code.
        :return: the parsed enum member or `None` if `value` is not valid :class:`Sex` value.
        """
        value = value.upper()
        if value == 'MALE':
            return Sex.MALE
        elif value == 'FEMALE':
            return Sex.FEMALE
        elif value == 'UNKNOWN':
            return Sex.UNKNOWN
        else:
            return None


class AnnotationReference(Identified):

    def __init__(self, identifier: TermId,
                 evidence_code: EvidenceCode):
        if not isinstance(identifier, TermId):
            raise ValueError(f'Identifier {identifier} is not a `TermId`')
        if not isinstance(evidence_code, EvidenceCode):
            raise ValueError(f'Evidence code {evidence_code} is not an `EvidenceCode`')
        self._identifier = identifier
        self._evidence_code = evidence_code

    @property
    def identifier(self) -> TermId:
        return self._identifier

    @property
    def evidence_code(self) -> EvidenceCode:
        return self._evidence_code

    def __eq__(self, other):
        return isinstance(other, AnnotationReference) \
            and self.identifier == other.identifier \
            and self.evidence_code == other.evidence_code

    def __hash__(self):
        return hash((self._identifier, self._evidence_code))

    def __str__(self):
        return f"AnnotationReference(" \
               f"identifier={self._identifier}, " \
               f"evidence_code={self._evidence_code})"

    def __repr__(self):
        return f"AnnotationReference(" \
               f"identifier={repr(self._identifier)}, " \
               f"evidence_code={repr(self._evidence_code)})"


class HpoDiseaseAnnotation(Identified, metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def ratio(self) -> Ratio:
        """
        :return: ratio representing a total number of the cohort members who displayed presence
        of the phenotypic feature represented by `HpoDiseaseAnnotation` at some point in their life.
        """
        pass

    @property
    @abc.abstractmethod
    def references(self) -> typing.List[AnnotationReference]:
        """
        :return: a list of `AnnotationReference`s that support presence/absence of the disease annotation
        """
        pass

    @property
    @abc.abstractmethod
    def modifiers(self) -> typing.List[TermId]:
        """
        :return: a list of disease annotation modifiers
        """
        pass

    def is_absent(self) -> bool:
        return self.ratio.is_zero()

    def is_present(self) -> bool:
        return not self.is_absent()

    def __str__(self):
        return f"HpoDiseaseAnnotation(" \
               f"identifier={self.identifier}, " \
               f"ratio={self.ratio}, " \
               f"references={self.references}, " \
               f"modifiers={self.modifiers})"


class HpoDisease(Identified, Named, metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def annotations(self) -> typing.Collection[HpoDiseaseAnnotation]:
        """
        :return a collection of all (present and absent) disease annotations.
        """
        pass

    @property
    @abc.abstractmethod
    def modes_of_inheritance(self) -> typing.Collection[TermId]:
        """
        :return: a collection of modes of inheritance associated with the disease.
        """
        pass

    def present_annotations(self) -> typing.Iterable[HpoDiseaseAnnotation]:
        """
        :return: an iterable over present disease features.
        """
        return filter(lambda a: a.is_present(), self.annotations)

    def absent_annotations(self) -> typing.Iterable[HpoDiseaseAnnotation]:
        """
        :return: an iterable over absent/excluded disease features.
        """
        return filter(lambda a: a.is_absent(), self.annotations)

    def __str__(self):
        return f"HpoDisease(" \
               f"identifier={self.identifier}, " \
               f"name={self.name}, " \
               f"n_annotations={len(self.annotations)})"


class HpoDiseases(Versioned, typing.Sized, metaclass=abc.ABCMeta):
    """
    A container for a set of :class:`HpoDisease`s that allows iteration over all diseases,
    knows about the number of diseases in the container, and supports retrieval of the disease by its identifier.
    """

    @property
    @abc.abstractmethod
    def diseases(self) -> typing.Iterable[HpoDisease]:
        """
        :return: an iterable over :class:`HpoDisease`s of the container.
        """
        pass

    @abc.abstractmethod
    def __getitem__(self, disease_id: CURIE_OR_TERM_ID) -> typing.Optional[HpoDisease]:
        """
        Get :class:`HpoDisease` based on given `disease_id` or `None` if the disease for the identifier is not present
        in the container.

        :param disease_id: a `str` or :class:`TermId` of the disease
        :return: :class:`HpoDisease` or `None`
        """
        pass

    @property
    def disease_ids(self) -> typing.Iterable[TermId]:
        """
        :return: an iterable over all disease IDs.
        """
        return map(lambda d: d.identifier, self.diseases)

    def __str__(self):
        return f"HpoDiseases(n_diseases={len(self)}, " \
               f"version={self.version})"



import abc
import logging
import typing

from hpotk.model import TermId
from ._api import OntologyGraph, NODE, OWL_THING
from ._csr_graph import SimpleCsrOntologyGraph
from .csr import CsrMatrixBuilder, ImmutableCsrMatrix

import numpy as np

# A newtype for stronger typing. We use these in `GraphFactory` below.
DirectedEdge = typing.Tuple[TermId, TermId]
GRAPH = typing.TypeVar('GRAPH', bound=OntologyGraph)


class GraphFactory(typing.Generic[GRAPH], metaclass=abc.ABCMeta):
    """
    Graph factory creates a graph from a list of `TermId` pairs
    """

    def __init__(self):
        self._logger = logging.getLogger(__name__)

    @abc.abstractmethod
    def create_graph(self, edge_list: typing.Sequence[DirectedEdge]) -> GRAPH:
        """
        Create graph from edge list.
        :param edge_list: a sequence of `DirectedEdge`s where the first item is the source
                          and the second item is the destination
        :return: the graph
        """
        pass


class CsrGraphFactory(GraphFactory[SimpleCsrOntologyGraph]):
    """
    A factory for creating `OntologyGraph` that is backed by a compressed sparse row (CSR) connectivity matrix.
    """

    def create_graph(self, edge_list: typing.Sequence[DirectedEdge]) -> SimpleCsrOntologyGraph:
        # Find root node
        self._logger.debug(f'Creating ontology graph from {len(edge_list)} edges')
        root, edge_list = _phenol_find_root(edge_list)
        self._logger.debug(f'Found root {root.value}')

        # Prepare node list
        nodes = _extract_nodes(edge_list)
        self._logger.debug(f'Extracted {len(nodes)} nodes')

        # Build connectivity matrix
        self._logger.debug(f'Building sparse adjacency matrix')
        node_to_idx = {node: idx for idx, node in enumerate(nodes)}
        builder = CsrMatrixBuilder(shape=(len(nodes), len(nodes)))

        for edge in edge_list:
            src_idx = node_to_idx[edge[0]]
            dest_idx = node_to_idx[edge[1]]
            builder[src_idx, dest_idx] = SimpleCsrOntologyGraph.PARENT_RELATIONSHIP_CODE
            builder[dest_idx, src_idx] = SimpleCsrOntologyGraph.CHILD_RELATIONSHIP_CODE

        self._logger.debug(f'Assembling the adjacency matrix')
        connectivity_matrix = ImmutableCsrMatrix(builder.row,
                                                 builder.col,
                                                 builder.data,
                                                 builder.shape,
                                                 dtype=int)
        # Assemble the ontology
        self._logger.debug(f'Finalizing the ontology graph')
        return SimpleCsrOntologyGraph(root, nodes, connectivity_matrix)


def _extract_and_sort_nodes(edge_list: typing.Sequence[DirectedEdge]):
    return np.sort(_extract_nodes(edge_list))


def _extract_nodes(edge_list: typing.Sequence[DirectedEdge]):
    nodes: typing.Set[TermId] = set()
    for edge in edge_list:
        nodes.add(edge[0])
        nodes.add(edge[1])
    return np.array(list(nodes), dtype=object)


def _phenol_find_root(edge_list: typing.Sequence[DirectedEdge]) -> typing.Tuple[NODE, typing.Sequence[DirectedEdge]]:
    """
    Find an ontology root candidate - the term that is parent of all elements using `DirectedEdge` that represents
    `src` -> `is_a` -> `dst` relationship.

    The method finds `TermId`s with no parents. There are 3 situations that can happen:
    - no `TermId` with no parents are found - we throw an error
    - one `TermId` with no parents is found, this is the root candidate
    - two or more `TermId`s with no parents are found, we add `OWL_THING` as the root term.
    :param edge_list:
    :return: a tuple with the root and possibly updated edge_list
    """
    root_candidate_set: typing.Set[TermId] = set()
    remove_mark_set: typing.Set[TermId] = set()
    for edge in edge_list:
        root_candidate_set.add(edge[1])
        remove_mark_set.add(edge[0])

    candidates = root_candidate_set.difference(remove_mark_set)
    if len(candidates) == 0:
        raise ValueError(f'No root candidate found')
    if len(candidates) == 1:
        return candidates.pop(), edge_list
    else:
        # No single root candidate, so create a new one and add it into the nodes and edges
        # As per suggestion https://github.com/monarch-initiative/phenol/issues/163#issuecomment-452880405
        # We'll use owl:Thing instead of ID:0000000 so as not to potentially conflict with an existing term id.
        # TODO - we need tests!
        edge_list_copy = [val for val in edge_list]
        for candidate in candidates:
            edge = (candidate, OWL_THING)
            edge_list_copy.append(edge)

        return OWL_THING, edge_list_copy



import abc
import typing

from ._term_id import TermId


class Identified(metaclass=abc.ABCMeta):
    """
    An entity that has a CURIE identifier.
    """

    @property
    @abc.abstractmethod
    def identifier(self) -> TermId:
        """
        :return: the identifier of the entity.
        """
        pass


class Named(metaclass=abc.ABCMeta):
    """
    An entity that has human-readable name or label.
    """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        :return: the human-readable name of the entity
        """
        pass


class Versioned(metaclass=abc.ABCMeta):
    """
    Base class for entities that may have version.
    """

    @property
    @abc.abstractmethod
    def version(self) -> typing.Optional[str]:
        """
        :return: version `str` or `None` if the version is not available.
        """
        pass
